from .src.load_data import LoadData
from .src.train_predacons import TrainPredacons
from .src.generate import Generate
from .src.predacons import (
    rollout,
    read_documents_from_directory,
    read_multiple_files,
    clean_text,
    train,
    trainer,
    generate_text,
    generate_output,
    generate_text_data_source_openai,
    generate_text_data_source_llm
)



import pandas as pd
import numpy as np
import re
from PyPDF2 import PdfReader
import os
import docx

class LoadData:
    def __read_pdf(file_path):
        with open(file_path, "rb") as file:
            pdf_reader = PdfReader(file)
            text = ""
            for page_num in range(len(pdf_reader.pages)):
                text += pdf_reader.pages[page_num].extract_text()
        return text

    def __read_word(file_path):
        doc = docx.Document(file_path)
        text = ""
        for paragraph in doc.paragraphs:
            text += paragraph.text + "\n"
        return text

    def __read_txt(file_path,encoding="utf-8"):
        with open(file_path, "r",encoding=encoding) as file:
            text = file.read()
        return text
    
    def clean_text(text):
        return re.sub(r'\n+', '\n', text).strip()
    
    def read_documents_from_directory(directory,encoding="utf-8"):
        combined_text = ""
        for filename in os.listdir(directory):
            file_path = os.path.join(directory, filename)
            if filename.endswith(".pdf"):
                combined_text += LoadData.__read_pdf(file_path)
            elif filename.endswith(".docx"):
                combined_text += LoadData.__read_word(file_path)
            elif filename.endswith(".txt"):
                combined_text += LoadData.__read_txt(file_path,encoding)
        return combined_text
    
    def read_multiple_files(file_paths):
        combined_text = ""
        for file_path in file_paths:
            if file_path.endswith(".pdf"):
                combined_text += LoadData.__read_pdf(file_path)
            elif file_path.endswith(".docx"):
                combined_text += LoadData.__read_word(file_path)
            elif file_path.endswith(".txt"):
                combined_text += LoadData.__read_txt(file_path)
        return combined_text


import openai
import pandas as pd
from .generate import Generate

class DataPreparation:
    def __generate_data_source_openai(client,gpt_model,prompt, existing_data, temperature=.5):
        
        messages=[
            {
                "role": "system",
                "content": f"You are generating data which will be used to train a machine learning model.\n\nYou will be given a high-level description of the model we want to train, and from that, you will generate data samples, each with a prompt/response pair.\n\nYou will do so in this format:\n```\nprompt\n-----------\n$prompt_goes_here\n-----------\n\nresponse\n-----------\n$response_goes_here\n-----------\n```\n\nOnly one prompt/response pair should be generated per turn.\n\nFor each turn, make the example slightly more complex than the last, while ensuring diversity.\n\nMake sure your samples are unique and diverse, yet high-quality and complex enough to train a well-performing model.\n\nHere is the type of model we want to train:\n`{prompt}`"
            }
        ]
        if len(existing_data) > 0:
            if len(existing_data) > 10:
                existing_data = random.sample(existing_data, 10)
            for example in existing_data:
                messages.append({
                    "role": "assistant",
                    "content": example
                })

        completion = client.chat.completions.create(
            model=gpt_model,
            messages=messages,
        )
        return(completion.choices[0].message.content)
    
    def __generate_data_source_llm(model_path, sequence, max_length,existing_data,trust_remote_code = False):
        messages=f"system : you are assistant and you are generating data which will be used to train a machine learning model.\n\nYou will be given a high-level description of the model we want to train, and from that, you will generate data samples, each with a prompt/response pair.\n\nYou will do so in this format:\n```\nprompt\n-----------\n$prompt_goes_here\n-----------\n\nresponse\n-----------\n$response_goes_here\n-----------\n```\n\nOnly one prompt/response pair should be generated per turn.\n\nFor each turn, make the example slightly more complex than the last, while ensuring diversity.\n\nMake sure your samples are unique and diverse, yet high-quality and complex enough to train a well-performing model.\n\nHere is the type of model we want to train:\n`{sequence}`"
            
        if len(existing_data) > 0:
            if len(existing_data) > 10:
                existing_data = random.sample(existing_data, 10)
            for example in existing_data:
                messages = messages + f" \n assistant : {example}"
        return Generate.generate_text(model_path, messages, max_length ,trust_remote_code = trust_remote_code)
        
    def __generate_text_data_source_openai(client,gpt_model,prompt,number_of_examples,temperature =0.5):
        prev_examples = []
        print(f'Generating data source for prompt: {prompt}')
        for i in range(int(number_of_examples)):
            example = DataPreparation.__generate_data_source_openai(client,gpt_model,prompt, prev_examples, temperature)
            prev_examples.append(example)

        training_data = "\n \n \n".join(prev_examples)
        return training_data
    
    def __generate_text_data_source_llm(model_path, sequence, max_length,number_of_examples,trust_remote_code=False):
        prev_examples = []
        print(f'Generating data source for prompt: {sequence}')
        for i in range(int(number_of_examples)):
            example=DataPreparation.__generate_data_source_llm(model_path, sequence, max_length,prev_examples,trust_remote_code = trust_remote_code)
            prev_examples.append(example)
        training_data = "\n \n \n".join(prev_examples)
        return training_data
    
    def generate_text_data_source_openai(client,gpt_model,prompt,number_of_examples,temperature =0.5):
        return DataPreparation.__generate_text_data_source_openai(client,gpt_model,prompt,number_of_examples,temperature)
    
    def generate_text_data_source_llm(model_path, sequence, max_length,number_of_examples,trust_remote_code=False):
        return DataPreparation.__generate_text_data_source_llm(model_path, sequence, max_length,number_of_examples,trust_remote_code=trust_remote_code)



from .load_data import LoadData
from .train_predacons import TrainPredacons
from .generate import Generate
from .data_preparation import DataPreparation

def rollout():
    print("Predacons rollout !!!")
    print("Predacons Version: v0.0.110")
    print("\nread_documents_from_directory -- Load data from directory")
    print("    directory -- Directory path")
    print("\nread_multiple_files -- Load data from multiple files")
    print("    file_paths -- list of File paths")
    print("\nclean_text -- Clean text")
    print("    text -- Text")
    print("\ntrain -- Train Predacons")
    print("    train_file_path -- Train file path")
    print("    model_name -- Model name")
    print("    output_dir -- Output directory")
    print("    overwrite_output_dir -- Overwrite output directory")
    print("    per_device_train_batch_size -- Per device train batch size")
    print("    num_train_epochs -- Number of train epochs")
    print("    save_steps -- Save steps")
    print("    trust_remote_code -- Trust remote code")
    print("\ntrainer -- returns trainer")
    print("    train_file_path -- Train file path")
    print("    model_name -- Model name")
    print("    output_dir -- Output directory")
    print("    overwrite_output_dir -- Overwrite output directory")
    print("    per_device_train_batch_size -- Per device train batch size")
    print("    num_train_epochs -- Number of train epochs")
    print("    save_steps -- Save steps")
    print("    trust_remote_code -- Trust remote code")
    print("\ngenerate_text -- Generate text")
    print("    model_path -- Model path")
    print("    sequence -- Sequence")
    print("    max_length -- Max length")
    print("    trust_remote_code -- Trust remote code")
    print("\ngenerate_output -- returns output and tokenizer")
    print("    model_path -- Model path")
    print("    sequence -- Sequence")
    print("    max_length -- Max length")
    print("    trust_remote_code -- Trust remote code")
    print("\ngenerate_text_data_source_openai -- Generate text data source using openai")
    print("    client -- openai client")
    print("    gpt_model -- GPT model used for generation")
    print("    prompt -- Prompt to generate data source")
    print("    number_of_examples -- Number of examples")
    print("    temperature -- Temperature (default 0.5)")
    print("\ngenerate_text_data_source_ll -- Generate text data source using local or hugging face llm")
    print("    model_path -- Model path or hugging face model name")
    print("    sequence -- Sequence (prompt)")
    print("    max_length -- Max length of the generated text")
    print("    number_of_examples -- Number of examples")
    print("    trust_remote_code -- Trust remote code")
    print("\nPredacons rollout !!!")

# Load data
def read_documents_from_directory(directory,encoding="utf-8"):
    return LoadData.read_documents_from_directory(directory,encoding)

def read_multiple_files(file_paths):
    return LoadData.read_multiple_files(file_paths)

def clean_text(text):
    return LoadData.clean_text(text)

# Train Predacons
def train(train_file_path,model_name,
          output_dir,
          overwrite_output_dir,
          per_device_train_batch_size,
          num_train_epochs,
          save_steps,
          trust_remote_code = False,
          resume_from_checkpoint=True):
    TrainPredacons.train(train_file_path,model_name,
          output_dir,
          overwrite_output_dir,
          per_device_train_batch_size,
          num_train_epochs,
          save_steps,
          trust_remote_code = trust_remote_code,
          resume_from_checkpoint=resume_from_checkpoint)

# get trainer
def trainer(train_file_path,model_name,
          output_dir,
          overwrite_output_dir,
          per_device_train_batch_size,
          num_train_epochs,
          save_steps,
          trust_remote_code = False):
    return TrainPredacons.trainer(train_file_path,model_name,
          output_dir,
          overwrite_output_dir,
          per_device_train_batch_size,
          num_train_epochs,
          save_steps,
          trust_remote_code = trust_remote_code)
    
# Generate text
def generate_text(model_path, sequence, max_length,trust_remote_code = False):
    return Generate.generate_text(model_path, sequence, max_length,trust_remote_code = trust_remote_code)

# Generate output
def generate_output(model_path, sequence, max_length,trust_remote_code = False):
    return Generate.generate_output(model_path, sequence, max_length,trust_remote_code = trust_remote_code)

# Data preparation
def generate_text_data_source_openai(client,gpt_model,prompt,number_of_examples,temperature =0.5):
    return DataPreparation.generate_text_data_source_openai(client,gpt_model,prompt,number_of_examples,temperature)

def generate_text_data_source_llm(model_path, sequence, max_length,number_of_examples,trust_remote_code=False):
    return DataPreparation.generate_text_data_source_llm(model_path, sequence, max_length,number_of_examples,trust_remote_code=trust_remote_code)

 




from transformers import TextDataset, DataCollatorForLanguageModeling
from transformers import AutoTokenizer,AutoModelForPreTraining,AutoModelForCausalLM
from transformers import Trainer, TrainingArguments

class TrainPredacons:

    def __load_dataset(file_path, tokenizer, block_size = 128):
        dataset = TextDataset(
            tokenizer = tokenizer,
            file_path = file_path,
            block_size = block_size,
        )
        return dataset
    
    def __load_data_collator(tokenizer, mlm = False):
        data_collator = DataCollatorForLanguageModeling(
            tokenizer=tokenizer, 
            mlm=mlm,
        )
        return data_collator
    
    def __trainer(train_file_path,model_name,
          output_dir,
          overwrite_output_dir,
          per_device_train_batch_size,
          num_train_epochs,
          save_steps,trust_remote_code = False):
        tokenizer = AutoTokenizer.from_pretrained(model_name)
        train_dataset = TrainPredacons.__load_dataset(train_file_path, tokenizer)
        data_collator = TrainPredacons.__load_data_collator(tokenizer)

        tokenizer.save_pretrained(output_dir)
        model = None
        try:
            model = AutoModelForPreTraining.from_pretrained(model_name,trust_remote_code=trust_remote_code)
        except:
            model = AutoModelForCausalLM.from_pretrained(model_name,trust_remote_code=trust_remote_code)

        model.save_pretrained(output_dir)

        training_args = TrainingArguments(
                output_dir=output_dir,
                overwrite_output_dir=overwrite_output_dir,
                per_device_train_batch_size=per_device_train_batch_size,
                num_train_epochs=num_train_epochs,
            )

        trainer = Trainer(
                model=model,
                args=training_args,
                data_collator=data_collator,
                train_dataset=train_dataset,
        )
        return trainer

    def __train(train_file_path,model_name,
          output_dir,
          overwrite_output_dir,
          per_device_train_batch_size,
          num_train_epochs,
          save_steps,trust_remote_code = False,resume_from_checkpoint=True):

        trainer = TrainPredacons.__trainer(train_file_path,model_name,
            output_dir,
            overwrite_output_dir,
            per_device_train_batch_size,
            num_train_epochs,
            save_steps,trust_remote_code = trust_remote_code)
        try:
            trainer.train(resume_from_checkpoint=resume_from_checkpoint)
        except:
            print('Failed to resume from checkpoint. training from scratch.')
            trainer.train()
        trainer.save_model()

    def trainer(train_file_path,model_name,
          output_dir,
          overwrite_output_dir,
          per_device_train_batch_size,
          num_train_epochs,
          save_steps,trust_remote_code = False):
        return TrainPredacons.__trainer(train_file_path,model_name,
            output_dir,
            overwrite_output_dir,
            per_device_train_batch_size,
            num_train_epochs,
            save_steps,trust_remote_code = trust_remote_code)
    
    def train(train_file_path,model_name,
          output_dir,
          overwrite_output_dir,
          per_device_train_batch_size,
          num_train_epochs,
          save_steps,trust_remote_code = False,resume_from_checkpoint = True):
        TrainPredacons.__train(train_file_path,model_name,output_dir,overwrite_output_dir,per_device_train_batch_size,num_train_epochs,save_steps,trust_remote_code = trust_remote_code,resume_from_checkpoint=resume_from_checkpoint)


from setuptools import find_packages, setup
import pathlib
with open("README.md", "r") as f:
    long_description = f.read()

HERE = pathlib.Path(__file__).parent

setup(
    name="predacons",
    version="0.0.113",
    description="A python library based on transformers for transfer learning",
    package_dir={"": "app"},
    packages=find_packages(where="app"),
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/shouryashashank/predacons",
    author="shouryashashank",
    author_email="shouryashashank@gmail.com",
    license="MIT",
    classifiers=[
        "License :: OSI Approved :: MIT License",   
        "Programming Language :: Python :: 3.10",
        "Operating System :: OS Independent",
    ],  
    install_requires=["bson >= 0.5.10",
                    "pandas >= 1.5.3",
                    "numpy >= 1.24.1",
                    "regex >= 2021.4.4",
                    "PyPDF2 >= 3.0.1",
                    "docx >= 0.2.4",
                    "python-docx >= 1.0.1",
                    "transformers >= 4.29.1",
                    "einops >= 0.7.0",
                    "openai >= 1.12.0"]
,
    extras_require={
        "dev": ["pytest>=7.0", "twine>=4.0.2"],
    },
    python_requires=">=3.10",
)



from transformers import AutoModelForPreTraining, AutoTokenizer,AutoModelForCausalLM

class Generate:
    def __load_model(model_path, trust_remote_code=False):
        model = None
        try:
            model = AutoModelForPreTraining.from_pretrained(model_path,trust_remote_code=trust_remote_code)
        except:
            model = AutoModelForCausalLM.from_pretrained(model_path,trust_remote_code=trust_remote_code)
        return model


    def __load_tokenizer(tokenizer_path):
        tokenizer = AutoTokenizer.from_pretrained(tokenizer_path)
        return tokenizer
    
    def __generate_output(model_path, sequence, max_length,trust_remote_code=False):
        model = Generate.__load_model(model_path,trust_remote_code=trust_remote_code)
        tokenizer = Generate.__load_tokenizer(model_path)
        ids = tokenizer.encode(f'{sequence}', return_tensors='pt')
        final_outputs = model.generate(
            ids,
            do_sample=True,
            max_length=max_length,
            pad_token_id=model.config.eos_token_id,
            top_k=50,
            top_p=0.95,
        )
        return final_outputs,tokenizer
    
    def __generate_text(model_path, sequence, max_length,trust_remote_code=False):
        final_outputs,tokenizer = Generate.__generate_output(model_path, sequence, max_length,trust_remote_code=trust_remote_code)
        
        print(tokenizer.decode(final_outputs[0], skip_special_tokens=True))
        return (tokenizer.decode(final_outputs[0], skip_special_tokens=True))

    def generate_output(model_path, sequence, max_length,trust_remote_code=False):
        return Generate.__generate_output(model_path, sequence, max_length,trust_remote_code=trust_remote_code)
    
    def generate_text(model_path, sequence, max_length,trust_remote_code=False):
        return Generate.__generate_text(model_path, sequence, max_length,trust_remote_code=trust_remote_code)